"""Core spy logic."""
import inspect
import functools
import warnings
from typing import Any, Dict, NamedTuple, Optional, Tuple, Type, Union, get_type_hints

from .spy_events import SpyInfo
from .warnings import IncorrectCallWarning


class _FROM_SOURCE:
    pass


FROM_SOURCE = _FROM_SOURCE()
"""Indicates a value that should be derived from the source object."""


class BoundArgs(NamedTuple):
    """Arguments bound to a spec."""

    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]


_DEFAULT_SPY_NAME = "unnamed"


class SpyCore:
    """Core spy logic for mimicking a given `source` object.

    Arguments:
        source: The source object the Spy is mimicking.
        name: The spec's name. If `None`, will be derived from `source`.
            Will fallback to a default value.
        module_name: The spec's module name. If left unspecified,
            will be derived from `source`. If explicitly set to `None`,
            a module name will not be used.
        is_async: If the spy should be configured to be an asynchronous callable.
            If `False` or unspecified, will derive from `source`.
    """

    def __init__(
        self,
        source: Optional[object],
        name: Optional[str],
        module_name: Union[str, _FROM_SOURCE, None] = FROM_SOURCE,
        is_async: bool = False,
    ) -> None:
        source = _resolve_source(source)

        self._source = source
        self._name = _get_name(source) if name is None else name
        self._module_name = (
            _get_module_name(source) if module_name is FROM_SOURCE else module_name
        )
        self._full_name = (
            f"{self._module_name}.{self._name}" if self._module_name else self._name
        )
        self._class_type = self._source if inspect.isclass(self._source) else None
        self._signature = _get_signature(source)
        self._is_async = is_async or _get_is_async(source)
        self._info = SpyInfo(id=id(self), name=self._name, is_async=self._is_async)

    @property
    def info(self) -> SpyInfo:
        """Get the spy's information object, for use in SpyLog events."""
        return self._info

    @property
    def full_name(self) -> str:
        """Get the full name of the spy, including module name, if available."""
        return self._full_name

    @property
    def signature(self) -> Optional[inspect.Signature]:
        """Get the spy's signature, if spy represents a callable."""
        return self._signature

    @property
    def class_type(self) -> Optional[Type[Any]]:
        """Get the spy's class type, if spy represents a class."""
        return self._class_type

    @property
    def is_async(self) -> bool:
        """Get whether the spy represents an asynchronous callable."""
        return self._is_async

    def bind_args(self, *args: Any, **kwargs: Any) -> BoundArgs:
        """Bind given args and kwargs to the Spec's signature, if possible.

        If no signature or unable to bind, will simply pass args and kwargs
        through without modification.
        """
        signature = self._signature

        if signature:
            try:
                bound_args = signature.bind(*args, **kwargs)
            except TypeError as e:
                # stacklevel: 4 ensures warning is linked to call location
                warnings.warn(IncorrectCallWarning(e), stacklevel=4)
            else:
                args = bound_args.args
                kwargs = bound_args.kwargs

        return BoundArgs(args=args, kwargs=kwargs)

    def create_child_core(self, name: str, is_async: bool) -> "SpyCore":
        """Create a child attribute, property, or method's SpyCore."""
        source = self._source
        child_name = f"{self._name}.{name}"
        child_source = None

        if inspect.isclass(source):
            # use type hints to get child spec for class attributes
            child_hint = _get_type_hints(source).get(name)
            # use inspect to get child spec for methods and properties
            child_source = inspect.getattr_static(source, name, child_hint)

            if isinstance(child_source, property):
                child_source = _get_type_hints(child_source.fget).get("return")

            elif isinstance(child_source, staticmethod):
                child_source = child_source.__func__

            else:
                if isinstance(child_source, classmethod):
                    child_source = child_source.__func__

                child_source = inspect.unwrap(child_source)

                if inspect.isfunction(child_source):
                    # consume the `self` argument of the method to ensure proper
                    # signature reporting by wrapping it in a partial
                    child_source = functools.partial(child_source, None)

        return SpyCore(
            source=child_source,
            name=child_name,
            module_name=self._module_name,
            is_async=is_async,
        )


def _resolve_source(source: Any) -> Any:
    """Resolve the source object, unwrapping any generic aliases."""
    origin = inspect.getattr_static(source, "__origin__", None)

    return origin if origin is not None else source


def _get_name(source: Any) -> str:
    """Get the name of a source object."""
    source_name = getattr(source, "__name__", None) if source is not None else None
    return source_name if isinstance(source_name, str) else _DEFAULT_SPY_NAME


def _get_module_name(source: Any) -> Optional[str]:
    """Get the name of a source object."""
    module_name = getattr(source, "__module__", None) if source is not None else None
    return module_name if isinstance(module_name, str) else None


def _get_signature(source: Any) -> Optional[inspect.Signature]:
    """Get the signature of a source object."""
    source = _get_callable_source(source)

    try:
        return inspect.signature(source, follow_wrapped=True)
    except (ValueError, TypeError):
        return None


def _get_is_async(source: Any) -> bool:
    """Get whether the source is an asynchronous callable."""
    source = _get_callable_source(source)

    # `iscoroutinefunction` does not work for `partial` on Python < 3.8
    if isinstance(source, functools.partial):
        source = source.func

    return inspect.iscoroutinefunction(source)


def _get_callable_source(source: Any) -> Any:
    """Return the source's callable, checking if a class has a __call__ method."""
    # check if spec source is a class with a __call__ method
    if inspect.isclass(source):
        call_method = inspect.getattr_static(source, "__call__", None)
        if inspect.isfunction(call_method):
            # consume the `self` argument of the method to ensure proper
            # signature reporting by wrapping it in a partial
            source = functools.partial(call_method, None)

    return source


def _get_type_hints(obj: Any) -> Dict[str, Any]:
    """Get type hints for an object, if possible.

    The builtin `typing.get_type_hints` may fail at runtime,
    e.g. if a type is subscriptable according to mypy but not
    according to Python.
    """
    try:
        return get_type_hints(obj)
    except Exception:
        return {}



"""Spy call handling."""
from typing import Any, NamedTuple, Optional

from .spy_log import SpyLog
from .context_managers import ContextWrapper
from .spy_events import SpyCall, SpyEvent
from .stub_store import MISSING, StubStore


class CallHandlerResult(NamedTuple):
    """A return value from a call."""

    value: Any


class CallHandler:
    """An interface to handle calls to spies."""

    def __init__(self, spy_log: SpyLog, stub_store: StubStore) -> None:
        """Initialize the CallHandler with access to SpyEvents and Stubs."""
        self._spy_log = spy_log
        self._stub_store = stub_store

    def handle(self, call: SpyEvent) -> Optional[CallHandlerResult]:
        """Handle a Spy's call, triggering stub behavior if necessary."""
        behavior = self._stub_store.get_by_call(call)
        self._spy_log.push(call)

        if behavior is None:
            return None

        if behavior.error:
            raise behavior.error

        return_value: Any

        if behavior.action:
            if isinstance(call.payload, SpyCall):
                return_value = behavior.action(
                    *call.payload.args,
                    **call.payload.kwargs,
                )
            else:
                return_value = behavior.action()

        elif behavior.context_value is not MISSING:
            return_value = ContextWrapper(behavior.context_value)

        else:
            return_value = behavior.return_value

        return CallHandlerResult(return_value)



"""Errors raised by Decoy.

See the [errors guide][] for more details.

[errors guide]: ../usage/errors-and-warnings/#errors
"""
from typing import Optional, Sequence

from .spy_events import SpyEvent, VerifyRehearsal
from .stringify import count, stringify_error_message


class MissingRehearsalError(ValueError):
    """An error raised when a Decoy method is called without rehearsal(s).

    This error is raised if you use [`when`][decoy.Decoy.when],
    [`verify`][decoy.Decoy.verify], or [`prop`][decoy.Decoy.prop] incorrectly
    in your tests. When using async/await, this error can be triggered if you
    forget to include `await` with your rehearsal.

    See the [MissingRehearsalError guide][] for more details.

    [MissingRehearsalError guide]: ../usage/errors-and-warnings/#missingrehearsalerror
    """

    def __init__(self) -> None:
        super().__init__("Rehearsal not found.")


class MockNotAsyncError(TypeError):
    """An error raised when an asynchronous function is used with a synchronous mock.

    This error is raised if you pass an `async def` function
    to a synchronous stub's `then_do` method.
    See the [MockNotAsyncError guide][] for more details.

    [MockNotAsyncError guide]: ../usage/errors-and-warnings/#mocknotasyncerror
    """


class VerifyError(AssertionError):
    """An error raised when actual calls do not match rehearsals given to `verify`.

    See [spying with verify][] for more details.

    [spying with verify]: ../usage/verify/

    Attributes:
        rehearsals: Rehearsals that were being verified.
        calls: Actual calls to the mock(s).
        times: The expected number of calls to the mock, if any.
    """

    rehearsals: Sequence[VerifyRehearsal]
    calls: Sequence[SpyEvent]
    times: Optional[int]

    def __init__(
        self,
        rehearsals: Sequence[VerifyRehearsal],
        calls: Sequence[SpyEvent],
        times: Optional[int],
    ) -> None:
        if times is not None:
            heading = f"Expected exactly {count(times, 'call')}:"
        elif len(rehearsals) == 1:
            heading = "Expected at least 1 call:"
        else:
            heading = "Expected call sequence:"

        message = stringify_error_message(
            heading=heading,
            rehearsals=rehearsals,
            calls=calls,
            include_calls=times is None or times == len(calls),
        )

        super().__init__(message)
        self.rehearsals = rehearsals
        self.calls = calls
        self.times = times



"""Spy objects.

Classes in this module are heavily inspired by the
[unittest.mock library](https://docs.python.org/3/library/unittest.mock.html).
"""
import inspect
from types import TracebackType
from typing import Any, ContextManager, Dict, Optional, Type, Union, cast, overload

from .call_handler import CallHandler
from .spy_core import SpyCore
from .spy_events import SpyCall, SpyEvent, SpyPropAccess, PropAccessType


class BaseSpy(ContextManager[Any]):
    """Spy object base class.

    - Pretends to be another class, if another class is given as a spec
    - Lazily constructs child spies when an attribute is accessed
    """

    _core: SpyCore
    _call_handler: CallHandler
    _spy_creator: "SpyCreator"
    _spy_children: Dict[str, "BaseSpy"]
    _spy_property_values: Dict[str, Any]

    def __init__(
        self,
        core: SpyCore,
        call_handler: CallHandler,
        spy_creator: "SpyCreator",
    ) -> None:
        """Initialize a BaseSpy from a call handler and an optional spec object."""
        super().__setattr__("_core", core)
        super().__setattr__("_call_handler", call_handler)
        super().__setattr__("_spy_creator", spy_creator)
        super().__setattr__("_spy_children", {})
        super().__setattr__("_spy_property_values", {})
        super().__setattr__("__signature__", self._core.signature)

    @property  # type: ignore[misc]
    def __class__(self) -> Any:
        """Ensure Spy can pass `instanceof` checks."""
        return self._core.class_type or type(self)

    def __enter__(self) -> Any:
        """Allow a spy to be used as a context manager."""
        enter_spy = self._get_or_create_child_spy("__enter__")
        return enter_spy()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        """Allow a spy to be used as a context manager."""
        exit_spy = self._get_or_create_child_spy("__exit__")
        return cast(Optional[bool], exit_spy(exc_type, exc_value, traceback))

    async def __aenter__(self) -> Any:
        """Allow a spy to be used as an async context manager."""
        enter_spy = self._get_or_create_child_spy("__aenter__", child_is_async=True)
        return await enter_spy()

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        """Allow a spy to be used as a context manager."""
        exit_spy = self._get_or_create_child_spy("__aexit__", child_is_async=True)
        return cast(Optional[bool], await exit_spy(exc_type, exc_value, traceback))

    def __repr__(self) -> str:
        """Get a helpful string representation of the spy."""
        return f"<Decoy mock `{self._core.full_name}`>"

    def __getattr__(self, name: str) -> Any:
        """Get a property of the spy, always returning a child spy."""
        # do not attempt to mock magic methods
        if name.startswith("__") and name.endswith("__"):
            return super().__getattribute__(name)

        return self._get_or_create_child_spy(name)

    def __setattr__(self, name: str, value: Any) -> None:
        """Set a property on the spy, recording the call."""
        event = SpyEvent(
            spy=self._core.info,
            payload=SpyPropAccess(
                prop_name=name,
                access_type=PropAccessType.SET,
                value=value,
            ),
        )
        self._call_handler.handle(event)
        self._spy_property_values[name] = value

    def __delattr__(self, name: str) -> None:
        """Delete a property on the spy, recording the call."""
        event = SpyEvent(
            spy=self._core.info,
            payload=SpyPropAccess(prop_name=name, access_type=PropAccessType.DELETE),
        )
        self._call_handler.handle(event)
        self._spy_property_values.pop(name, None)

    def _get_or_create_child_spy(self, name: str, child_is_async: bool = False) -> Any:
        """Lazily construct a child spy, basing it on type hints if available."""
        # check for any stubbed behaviors for property getter
        get_result = self._call_handler.handle(
            SpyEvent(
                spy=self._core.info,
                payload=SpyPropAccess(
                    prop_name=name,
                    access_type=PropAccessType.GET,
                ),
            )
        )

        if get_result:
            return get_result.value

        if name in self._spy_property_values:
            return self._spy_property_values[name]

        # return previously constructed (and cached) child spies
        if name in self._spy_children:
            return self._spy_children[name]

        child_core = self._core.create_child_core(name=name, is_async=child_is_async)
        child_spy = self._spy_creator.create(core=child_core)
        self._spy_children[name] = child_spy

        return child_spy

    def _call(self, *args: Any, **kwargs: Any) -> Any:
        bound_args, bound_kwargs = self._core.bind_args(*args, **kwargs)
        call = SpyEvent(
            spy=self._core.info,
            payload=SpyCall(
                args=bound_args,
                kwargs=bound_kwargs,
            ),
        )

        result = self._call_handler.handle(call)
        return result.value if result else None


class AsyncSpy(BaseSpy):
    """An object that records all async. calls made to itself and its children."""

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Handle a call to the spy asynchronously."""
        result = self._call(*args, **kwargs)
        return (await result) if inspect.iscoroutine(result) else result


class Spy(BaseSpy):
    """An object that records all calls made to itself and its children."""

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Handle a call to the spy."""
        return self._call(*args, **kwargs)


AnySpy = Union[AsyncSpy, Spy]


class SpyCreator:
    """Spy factory."""

    def __init__(self, call_handler: CallHandler) -> None:
        self._call_handler = call_handler

    @overload
    def create(self, *, core: SpyCore) -> AnySpy:
        ...

    @overload
    def create(
        self, *, spec: Optional[object], name: Optional[str], is_async: bool
    ) -> AnySpy:
        ...

    def create(
        self,
        *,
        core: Optional[SpyCore] = None,
        spec: Optional[object] = None,
        name: Optional[str] = None,
        is_async: bool = False,
    ) -> AnySpy:
        """Create a Spy from a spec.

        Functions and classes passed to `spec` will be inspected (and have any type
        annotations inspected) to ensure `AsyncSpy`'s are returned where necessary.
        """
        if not isinstance(core, SpyCore):
            core = SpyCore(source=spec, name=name, is_async=is_async)

        spy_cls: Type[AnySpy] = AsyncSpy if core.is_async else Spy

        return spy_cls(
            core=core,
            spy_creator=self,
            call_handler=self._call_handler,
        )



"""Stub creation and storage."""
from typing import Any, Callable, List, NamedTuple, Optional, Union

from .spy_events import SpyEvent, WhenRehearsal, match_event


class _MISSING:
    pass


MISSING = _MISSING()
"""Value not specified sentinel.

Used when `None` could be a valid value,
so `Optional` would be inappropriate.
"""


class StubBehavior(NamedTuple):
    """A recorded stub behavior."""

    return_value: Optional[Any] = None
    context_value: Union[_MISSING, Any] = MISSING
    error: Optional[Exception] = None
    action: Optional[Callable[..., Any]] = None
    once: bool = False


class StubEntry(NamedTuple):
    """An entry in the StubStore for later behavior lookup."""

    rehearsal: WhenRehearsal
    behavior: StubBehavior


class StubStore:
    """Stored stub behaviors."""

    def __init__(self) -> None:
        """Initialize a StubStore with an empty stubbings list."""
        self._stubs: List[StubEntry] = []

    def add(
        self,
        rehearsal: WhenRehearsal,
        behavior: StubBehavior,
    ) -> None:
        """Create and add a new StubBehavior to the store."""
        self._stubs.append(StubEntry(rehearsal=rehearsal, behavior=behavior))

    def get_by_call(self, call: SpyEvent) -> Optional[StubBehavior]:
        """Get the latest StubBehavior matching this call."""
        reversed_indices = range(len(self._stubs) - 1, -1, -1)

        for i in reversed_indices:
            stub = self._stubs[i]

            if match_event(call, stub.rehearsal):
                if stub.behavior.once:
                    self._stubs.pop(i)

                return stub.behavior

        return None

    def clear(self) -> None:
        """Remove all stored Stubs."""
        self._stubs.clear()



"""Wrappers around contextlib types and fallbacks."""
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    ContextManager,
    Generic,
    TypeVar,
)

# TODO(mc, 2022-03-14): drop support for Python 3.6 in Decoy v2
# Python 3.6 does not have async generator context managers
if TYPE_CHECKING:
    from contextlib import (  # type: ignore[attr-defined]
        _GeneratorContextManager as GeneratorContextManager,
        _AsyncGeneratorContextManager as AsyncGeneratorContextManager,
    )

_EnterT = TypeVar("_EnterT")


class ContextWrapper(
    ContextManager[_EnterT],
    AsyncContextManager[_EnterT],
    Generic[_EnterT],
):
    """A simple, do-nothing ContextManager that wraps a given value.

    Adapted from `contextlib.nullcontext` to ensure support across
    all Python versions.
    """

    def __init__(self, enter_result: _EnterT) -> None:
        self._enter_result = enter_result

    def __enter__(self) -> _EnterT:
        """Return the wrapped value."""
        return self._enter_result

    def __exit__(self, *args: Any, **kwargs: Any) -> Any:
        """No-op on exit."""
        pass

    async def __aenter__(self) -> _EnterT:
        """Return the wrapped value."""
        return self._enter_result

    async def __aexit__(self, *args: Any, **kwargs: Any) -> Any:
        """No-op on exit."""
        pass


__all__ = [
    "AsyncContextManager",
    "AsyncGeneratorContextManager",
    "ContextManager",
    "GeneratorContextManager",
    "ContextWrapper",
]



"""Spy interaction event value objects."""
import enum
from typing import Any, Dict, NamedTuple, Optional, Tuple, Union


class PropAccessType(str, enum.Enum):
    """Property access type."""

    GET = "get"
    SET = "set"
    DELETE = "delete"


class SpyInfo(NamedTuple):
    """Spy information and configuration."""

    id: int
    name: str
    is_async: bool


class SpyCall(NamedTuple):
    """Spy event payload representing a call to a function or method."""

    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]
    ignore_extra_args: bool = False


class SpyPropAccess(NamedTuple):
    """Spy event payload representing a get/set/delete of a property."""

    prop_name: str
    access_type: PropAccessType
    value: Optional[Any] = None


class SpyEvent(NamedTuple):
    """An interaction with a spy by the code under test."""

    spy: SpyInfo
    payload: Union[SpyCall, SpyPropAccess]


class WhenRehearsal(NamedTuple):
    """A spy interaction that has been used as a rehearsal for `when`."""

    spy: SpyInfo
    payload: Union[SpyCall, SpyPropAccess]


class VerifyRehearsal(NamedTuple):
    """A spy interaction that has been used as a rehearsal for `verify`."""

    spy: SpyInfo
    payload: Union[SpyCall, SpyPropAccess]


class PropRehearsal(NamedTuple):
    """A spy interaction that has been used as a rehearsal for `prop`."""

    spy: SpyInfo
    payload: SpyPropAccess


SpyRehearsal = Union[WhenRehearsal, VerifyRehearsal, PropRehearsal]

AnySpyEvent = Union[SpyEvent, SpyRehearsal]


def match_event(event: AnySpyEvent, rehearsal: SpyRehearsal) -> bool:
    """Check if a call matches a given rehearsal."""
    if event.spy != rehearsal.spy:
        return False

    if (
        isinstance(event.payload, SpyCall)
        and isinstance(rehearsal.payload, SpyCall)
        and rehearsal.payload.ignore_extra_args
    ):
        call = event.payload
        rehearsed_call = rehearsal.payload

        try:
            args_match = all(
                call.args[i] == value for i, value in enumerate(rehearsed_call.args)
            )
            kwargs_match = all(
                call.kwargs[key] == value
                for key, value in rehearsed_call.kwargs.items()
            )

            return args_match and kwargs_match

        except (IndexError, KeyError):
            return False

    return event.payload == rehearsal.payload



"""Matcher helpers.

A "matcher" is a class with an `__eq__` method defined. Use them anywhere
in your test where you would use an actual value for equality (`==`) comparision.

Matchers help you loosen assertions where strict adherence to an exact value
is not relevent to what you're trying to test.

Example:
    ```python
    from decoy import Decoy, matchers

    # ...

    def test_logger_called(decoy: Decoy):
        # ...
        decoy.verify(
            logger.log(msg=matchers.StringMatching("hello"))
        )
    ```

Note:
    Identity comparisons (`is`) will not work with matchers. Decoy only uses
    equality comparisons (`==`) for stubbing and verification.
"""
from re import compile as compile_re
from typing import cast, Any, List, Mapping, Optional, Pattern, Type, TypeVar


__all__ = [
    "Anything",
    "IsA",
    "IsNot",
    "StringMatching",
    "ErrorMatching",
    "Captor",
]


class _Anything:
    def __eq__(self, target: object) -> bool:
        """Return true if target is not None."""
        return target is not None

    def __repr__(self) -> str:
        """Return a string representation of the matcher."""
        return "<Anything>"


def Anything() -> Any:
    """Match anything except None.

    Example:
        ```python
        assert "foobar" == Anything()
        assert None != Anything()
        ```
    """
    return _Anything()


class _IsA:
    _match_type: type
    _attributes: Optional[Mapping[str, Any]]

    def __init__(
        self,
        match_type: type,
        attributes: Optional[Mapping[str, Any]] = None,
    ) -> None:
        """Initialize the matcher with a type and optional attributes."""
        self._match_type = match_type
        self._attributes = attributes

    def __eq__(self, target: object) -> bool:
        """Return true if target is the correct type and matches attributes."""
        matches_type = isinstance(target, self._match_type)
        matches_attrs = (
            target == HasAttributes(self._attributes) if self._attributes else True
        )

        return matches_type and matches_attrs

    def __repr__(self) -> str:
        """Return a string representation of the matcher."""
        if self._attributes is None:
            return f"<IsA {self._match_type.__name__}>"
        else:
            return f"<IsA {self._match_type.__name__} {repr(self._attributes)}>"


def IsA(match_type: type, attributes: Optional[Mapping[str, Any]] = None) -> Any:
    """Match anything that satisfies the passed in type.

    Arguments:
        match_type: Type to match.
        attributes: Optional set of attributes to match

    Example:
        ```python
        assert "foobar" == IsA(str)
        assert datetime.now() == IsA(datetime)
        assert 42 == IsA(int)

        @dataclass
        class HelloWorld:
            hello: str = "world"
            goodby: str = "so long"

        assert HelloWorld() == IsA(HelloWorld, {"hello": "world"})
        ```
    """
    return _IsA(match_type, attributes)


class _IsNot:
    _reject_value: object

    def __init__(self, value: object) -> None:
        """Initialize the matcher with the value to reject."""
        self._reject_value = value

    def __eq__(self, target: object) -> bool:
        """Return true if target is not self._reject_value."""
        return target != self._reject_value

    def __repr__(self) -> str:
        """Return a string representation of the matcher."""
        return f"<IsNot {repr(self._reject_value)}>"


def IsNot(value: object) -> Any:
    """Match anything that isn't the passed in value.

    Arguments:
        value: Value to check against.

    Example:
        ```python
        assert "foobar" == IsNot("bazquux")
        assert 42 == IsNot("the question")
        assert 1 != IsNot(1)
        ```
    """
    return _IsNot(value)


class _HasAttributes:
    _attributes: Mapping[str, Any]

    def __init__(self, attributes: Mapping[str, Any]) -> None:
        self._attributes = attributes

    def __eq__(self, target: object) -> bool:
        """Return true if target matches all given attributes."""
        is_match = True
        for attr_name, value in self._attributes.items():
            if is_match:
                is_match = (
                    hasattr(target, attr_name) and getattr(target, attr_name) == value
                )

        return is_match

    def __repr__(self) -> str:
        """Return a string representation of the matcher."""
        return f"<HasAttributes {repr(self._attributes)}>"


def HasAttributes(attributes: Mapping[str, Any]) -> Any:
    """Match anything with the passed in attributes.

    Arguments:
        attributes: Attribute values to check.

    Example:
        ```python
        @dataclass
        class HelloWorld:
            hello: str = "world"
            goodby: str = "so long"

        assert HelloWorld() == matchers.HasAttributes({"hello": "world"})
        ```
    """
    return _HasAttributes(attributes)


class _DictMatching:
    _values: Mapping[str, Any]

    def __init__(self, values: Mapping[str, Any]) -> None:
        self._values = values

    def __eq__(self, target: object) -> bool:
        """Return true if target matches all given keys/values."""
        is_match = True

        for key, value in self._values.items():
            if is_match:
                try:
                    is_match = key in target and target[key] == value  # type: ignore[index,operator]  # noqa: E501
                except TypeError:
                    is_match = False

        return is_match

    def __repr__(self) -> str:
        """Return a string representation of the matcher."""
        return f"<DictMatching {repr(self._values)}>"


def DictMatching(values: Mapping[str, Any]) -> Any:
    """Match any dictionary with the passed in keys / values.

    Arguments:
        values: Keys and values to check.

    Example:
        ```python
        value = {"hello": "world", "goodbye": "so long"}
        assert value == matchers.DictMatching({"hello": "world"})
        ```
    """
    return _DictMatching(values)


class _StringMatching:
    _pattern: Pattern[str]

    def __init__(self, match: str) -> None:
        """Initialize the matcher with the pattern to match."""
        self._pattern = compile_re(match)

    def __eq__(self, target: object) -> bool:
        """Return true if target is not self._reject_value."""
        return isinstance(target, str) and self._pattern.search(target) is not None

    def __repr__(self) -> str:
        """Return a string representation of the matcher."""
        return f"<StringMatching {repr(self._pattern.pattern)}>"


def StringMatching(match: str) -> str:
    """Match any string matching the passed in pattern.

    Arguments:
        match: Pattern to check against; will be compiled into an re.Pattern.

    Example:
        ```python
        assert "foobar" == StringMatching("bar")
        assert "foobar" != StringMatching("^bar")
        ```
    """
    return cast(str, _StringMatching(match))


class _ErrorMatching:
    _error_type: Type[BaseException]
    _string_matcher: Optional[_StringMatching]

    def __init__(self, error: Type[BaseException], match: Optional[str] = None) -> None:
        """Initialize with the Exception type and optional message matcher."""
        self._error_type = error
        self._string_matcher = _StringMatching(match) if match is not None else None

    def __eq__(self, target: object) -> bool:
        """Return true if target is not self._reject_value."""
        error_match = type(target) == self._error_type
        message_match = (
            str(target) == self._string_matcher
            if self._string_matcher is not None
            else True
        )

        return error_match and message_match

    def __repr__(self) -> str:
        """Return a string representation of the matcher."""
        return (
            f"<ErrorMatching {self._error_type.__name__} match={self._string_matcher}>"
        )


ErrorT = TypeVar("ErrorT", bound=BaseException)


def ErrorMatching(error: Type[ErrorT], match: Optional[str] = None) -> ErrorT:
    """Match any error matching an Exception type and optional message matcher.

    Arguments:
        error: Exception type to match against.
        match: Pattern to check against; will be compiled into an re.Pattern.

    Example:
        ```python
        assert ValueError("oh no!") == ErrorMatching(ValueError)
        assert ValueError("oh no!") == ErrorMatching(ValueError, match="no")
        ```
    """
    return cast(ErrorT, _ErrorMatching(error, match))


class _Captor:
    def __init__(self) -> None:
        self._values: List[Any] = []

    def __eq__(self, target: object) -> bool:
        """Capture compared value, always returning True."""
        self._values.append(target)
        return True

    def __repr__(self) -> str:
        """Return a string representation of the matcher."""
        return "<Captor>"

    @property
    def value(self) -> Any:
        """Get the captured value.

        Raises:
            AssertionError: if no value was captured.
        """
        if len(self._values) == 0:
            raise AssertionError("No value captured by captor.")

        return self._values[-1]

    @property
    def values(self) -> List[Any]:
        """Get all captured values."""
        return self._values


def Captor() -> Any:
    """Match anything, capturing its value.

    The last captured value will be set to `captor.value`. All captured
    values will be placed in the `captor.values` list, which can be
    helpful if a captor needs to be triggered multiple times.

    Example:
        ```python
        captor = Captor()
        assert "foobar" == captor
        print(captor.value)  # "foobar"
        print(captor.values)  # ["foobar"]
        ```
    """
    return _Captor()



"""Message string generation."""
import os
from typing import Sequence

from .spy_events import AnySpyEvent, SpyCall, SpyEvent, SpyRehearsal, PropAccessType


def stringify_call(event: AnySpyEvent) -> str:
    """Stringify the call to something human readable.

    ```python
    SpyEvent(spy=SpyInfo(id=42, name="some_func"),
        payload=SpyCall(args=(1,), kwargs={"foo": False})
    )
    ```

    ...would stringify as `"some_func(1, foo=False)"`
    """
    spy, payload = event

    if not isinstance(payload, SpyCall):
        full_prop_name = f"{spy.name}.{payload.prop_name}"

        if payload.access_type == PropAccessType.SET:
            return f"{full_prop_name} = {payload.value}"
        elif payload.access_type == PropAccessType.DELETE:
            return f"del {full_prop_name}"

        return full_prop_name

    else:
        args_list = [repr(arg) for arg in payload.args]
        kwargs_list = [f"{key}={repr(val)}" for key, val in payload.kwargs.items()]
        extra_args_msg = (
            " - ignoring unspecified arguments" if payload.ignore_extra_args else ""
        )
        return f"{spy.name}({', '.join(args_list + kwargs_list)}){extra_args_msg}"


def stringify_call_list(calls: Sequence[AnySpyEvent]) -> str:
    """Stringify a sequence of calls into an ordered list."""
    return os.linesep.join(
        f"{i + 1}.\t{stringify_call(call)}" for i, call in enumerate(calls)
    )


def count(count: int, noun: str) -> str:
    """Count a given noun, pluralizing if necessary."""
    return f"{count} {noun}{'s' if count != 1 else ''}"


def join_lines(*lines: str) -> str:
    """Join a list of lines with newline characters."""
    return os.linesep.join(lines).strip()


def stringify_error_message(
    heading: str,
    rehearsals: Sequence[SpyRehearsal],
    calls: Sequence[SpyEvent],
    include_calls: bool = True,
) -> str:
    """Stringify an error message about a rehearsals to calls comparison."""
    return join_lines(
        heading,
        stringify_call_list(rehearsals),
        (
            f"Found {count(len(calls), 'call')}"
            f"{'.' if len(calls) == 0 or not include_calls else ':'}"
        ),
        stringify_call_list(calls) if include_calls else "",
    )



"""Decoy stubbing and spying library."""
from warnings import warn
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Generic,
    Optional,
    Union,
    cast,
    overload,
)

from . import errors, matchers, warnings
from .core import DecoyCore, StubCore, PropCore
from .types import ClassT, ContextValueT, FuncT, ReturnT

# TODO(mc, 2022-03-14): drop support for Python 3.6 in Decoy v2
# Python 3.6 does not have async generator context managers
if TYPE_CHECKING:
    from .context_managers import (
        ContextManager,
        AsyncContextManager,
        GeneratorContextManager,
        AsyncGeneratorContextManager,
    )

# ensure decoy does not pollute pytest tracebacks
__tracebackhide__ = True


class Decoy:
    """Decoy mock factory and state container.

    You should create a new Decoy instance before each test and call
    [`reset`][decoy.Decoy.reset] after each test. If you use the
    [`decoy` pytest fixture][decoy.pytest_plugin.decoy], this is done
    automatically. See the [setup guide](../#setup) for more details.

    Example:
        ```python
        decoy = Decoy()

        # test your subject
        ...

        decoy.reset()
        ```
    """

    def __init__(self) -> None:
        self._core = DecoyCore()

    @overload
    def mock(self, *, cls: Callable[..., ClassT]) -> ClassT:
        ...

    @overload
    def mock(self, *, func: FuncT) -> FuncT:
        ...

    # TODO(mc, 2021-11-14): make `name` required for specless mocks in v2.0
    @overload
    def mock(self, *, name: Optional[str] = None, is_async: bool = False) -> Any:
        ...

    def mock(
        self,
        *,
        cls: Optional[Any] = None,
        func: Optional[Any] = None,
        name: Optional[str] = None,
        is_async: bool = False,
    ) -> Any:
        """Create a mock. See the [mock creation guide] for more details.

        [mock creation guide]: ../usage/create/

        Arguments:
            cls: A class definition that the mock should imitate.
            func: A function definition the mock should imitate.
            name: A name to use for the mock. If you do not use
                `cls` or `func`, you should add a `name`.
            is_async: Force the returned spy to be asynchronous. This argument
                only applies if you don't use `cls` nor `func`.

        Returns:
            A spy typecast as the object it's imitating, if any.

        Example:
            ```python
            def test_get_something(decoy: Decoy):
                db = decoy.mock(cls=Database)
                # ...
            ```
        """
        spec = cls or func
        return self._core.mock(spec=spec, name=name, is_async=is_async)

    def create_decoy(
        self,
        spec: Callable[..., ClassT],
        *,
        is_async: bool = False,
    ) -> ClassT:
        """Create a class mock for `spec`.

        !!! warning "Deprecated since v1.6.0"
            Use [`mock`][decoy.Decoy.mock] with the `cls` parameter, instead.
        """
        warn(
            "decoy.create_decoy is deprecated; use decoy.mock(cls=...) instead.",
            DeprecationWarning,
        )

        spy = self._core.mock(spec=spec, is_async=is_async)
        return cast(ClassT, spy)

    def create_decoy_func(
        self,
        spec: Optional[FuncT] = None,
        *,
        is_async: bool = False,
    ) -> FuncT:
        """Create a function mock for `spec`.

        !!! warning "Deprecated since v1.6.0"
            Use [`mock`][decoy.Decoy.mock] with the `func` parameter, instead.
        """
        warn(
            "decoy.create_decoy_func is deprecated; use decoy.mock(func=...) instead.",
            DeprecationWarning,
        )

        spy = self._core.mock(spec=spec, is_async=is_async)
        return cast(FuncT, spy)

    def when(
        self,
        _rehearsal_result: ReturnT,
        *,
        ignore_extra_args: bool = False,
    ) -> "Stub[ReturnT]":
        """Create a [`Stub`][decoy.Stub] configuration using a rehearsal call.

        See [stubbing usage guide](../usage/when/) for more details.

        Arguments:
            _rehearsal_result: The return value of a rehearsal, used for typechecking.
            ignore_extra_args: Allow the rehearsal to specify fewer arguments than
                the actual call. Decoy will compare and match any given arguments,
                ignoring unspecified arguments.

        Returns:
            A stub to configure using [`then_return`][decoy.Stub.then_return],
            [`then_raise`][decoy.Stub.then_raise], [`then_do`][decoy.Stub.then_do],
            or [`then_enter_with`][decoy.Stub.then_enter_with].

        Example:
            ```python
            db = decoy.mock(cls=Database)
            decoy.when(db.exists("some-id")).then_return(True)
            ```

        Note:
            The "rehearsal" is an actual call to the test fake. Because the
            call is written inside `when`, Decoy is able to infer that the call
            is a rehearsal for stub configuration purposes rather than a call
            from the code-under-test.
        """
        stub_core = self._core.when(
            _rehearsal_result,
            ignore_extra_args=ignore_extra_args,
        )
        return Stub(core=stub_core)

    def verify(
        self,
        *_rehearsal_results: Any,
        times: Optional[int] = None,
        ignore_extra_args: bool = False,
    ) -> None:
        """Verify a mock was called using one or more rehearsals.

        See [verification usage guide](../usage/verify/) for more details.

        Arguments:
            _rehearsal_results: The return value of rehearsals, unused except
                to determine how many rehearsals to verify.
            times: How many times the call should appear. If `times` is specified,
                the call count must match exactly, otherwise the call must appear
                at least once. The `times` argument must be used with exactly one
                rehearsal.
            ignore_extra_args: Allow the rehearsal to specify fewer arguments than
                the actual call. Decoy will compare and match any given arguments,
                ignoring unspecified arguments.

        Raises:
            VerifyError: The verification was not satisfied.

        Example:
            ```python
            def test_create_something(decoy: Decoy):
                gen_id = decoy.mock(func=generate_unique_id)

                # ...

                decoy.verify(gen_id("model-prefix_"))
            ```

        Note:
            A "rehearsal" is an actual call to the test fake. The fact that
            the call is written inside `verify` is purely for typechecking and
            API sugar. Decoy will pop the last call(s) to _any_ fake off its
            call stack, which will end up being the call inside `verify`.
        """
        self._core.verify(
            *_rehearsal_results,
            times=times,
            ignore_extra_args=ignore_extra_args,
        )

    def prop(self, _rehearsal_result: ReturnT) -> "Prop[ReturnT]":
        """Create property setter and deleter rehearsals.

        See [property mocking guide](../advanced/properties/) for more details.

        Arguments:
            _rehearsal_result: The property to mock, for typechecking.

        Returns:
            A prop rehearser on which you can call [`set`][decoy.Prop.set] or
            [`delete`][decoy.Prop.delete] to create property rehearsals.
        """
        prop_core = self._core.prop(_rehearsal_result)
        return Prop(core=prop_core)

    def reset(self) -> None:
        """Reset all mock state.

        This method should be called after every test to ensure spies and stubs
        don't leak between tests. The [`decoy`][decoy.pytest_plugin.decoy] fixture
        provided by the pytest plugin will call `reset` automatically.

        The `reset` method may also trigger warnings if Decoy detects any questionable
        mock usage. See [decoy.warnings][] for more details.
        """
        self._core.reset()


class Stub(Generic[ReturnT]):
    """A rehearsed Stub that can be used to configure mock behaviors.

    See [stubbing usage guide](../usage/when/) for more details.
    """

    def __init__(self, core: StubCore) -> None:
        self._core = core

    def then_return(self, *values: ReturnT) -> None:
        """Configure the stub to return value(s).

        Arguments:
            *values: Zero or more return values. Multiple values will result
                     in different return values for subsequent calls, with the
                     last value latching in once all other values have returned.
        """
        self._core.then_return(*values)

    def then_raise(self, error: Exception) -> None:
        """Configure the stub to raise an error.

        Arguments:
            error: The error to raise.

        Note:
            Setting a stub to raise will prevent you from writing new
            rehearsals, because they will raise. If you need to make more calls
            to [`when`][decoy.Decoy.when], you'll need to wrap your rehearsal
            in a `try`.
        """
        self._core.then_raise(error)

    def then_do(
        self,
        action: Callable[..., Union[ReturnT, Coroutine[Any, Any, ReturnT]]],
    ) -> None:
        """Configure the stub to trigger an action.

        Arguments:
            action: The function to call. Called with whatever arguments
                are actually passed to the stub. May be an `async def`
                function if the mock is also asynchronous.

        Raises:
            MockNotAsyncError: `action` was an `async def` function,
                but the mock is synchronous.
        """
        self._core.then_do(action)

    @overload
    def then_enter_with(
        self: "Stub[ContextManager[ContextValueT]]",
        value: ContextValueT,
    ) -> None:
        ...

    @overload
    def then_enter_with(
        self: "Stub[AsyncContextManager[ContextValueT]]",
        value: ContextValueT,
    ) -> None:
        ...

    @overload
    def then_enter_with(
        self: "Stub[GeneratorContextManager[ContextValueT]]",
        value: ContextValueT,
    ) -> None:
        ...

    @overload
    def then_enter_with(
        self: "Stub[AsyncGeneratorContextManager[ContextValueT]]",
        value: ContextValueT,
    ) -> None:
        ...

    def then_enter_with(
        self: Union[
            "Stub[ContextManager[ContextValueT]]",
            "Stub[GeneratorContextManager[ContextValueT]]",
            "Stub[AsyncContextManager[ContextValueT]]",
            "Stub[AsyncGeneratorContextManager[ContextValueT]]",
        ],
        value: ContextValueT,
    ) -> None:
        """Configure the stub to return a value wrapped in a context manager.

        The wrapping context manager is compatible with both the synchronous and
        asynchronous context manager interfaces.

        See the [context manager usage guide](../advanced/context-managers/)
        for more details.

        Arguments:
            value: A return value to wrap in a ContextManager.
        """
        self._core.then_enter_with(value)


class Prop(Generic[ReturnT]):
    """Rehearsal creator for mocking property setters and deleters.

    See [property mocking guide](../advanced/properties/) for more details.
    """

    def __init__(self, core: PropCore) -> None:
        self._core = core

    def set(self, value: ReturnT) -> None:
        """Create a property setter rehearsal.

        By wrapping `set` in a call to [`when`][decoy.Decoy.when] or
        [`verify`][decoy.Decoy.verify], you can stub or verify a call
        to a property setter.

        Arguments:
            value: The value

        Example:
            ```python
            some_obj = decoy.mock()
            some_obj.prop = 42
            decoy.verify(decoy.prop(some_obj.prop).set(42))
            ```
        """
        self._core.set(value)

    def delete(self) -> None:
        """Create a property deleter rehearsal.

        By wrapping `delete` in a call to [`when`][decoy.Decoy.when] or
        [`verify`][decoy.Decoy.verify], you can stub or verify a call
        to a property deleter.


        Example:
            ```python
            some_obj = decoy.mock()
            del some_obj.prop
            decoy.verify(decoy.prop(some_obj.prop).delete())
            ```
        """
        self._core.delete()


__all__ = ["Decoy", "Stub", "Prop", "matchers", "warnings", "errors"]



"""Pytest plugin to setup and teardown a Decoy instance.

The plugin will be registered with pytest when you install Decoy. It adds a
fixture without modifying any other pytest behavior. Its usage is optional
but highly recommended.
"""
from typing import Iterable

import pytest

from decoy import Decoy


@pytest.fixture()
def decoy() -> Iterable[Decoy]:
    """Get a [decoy.Decoy][] container and [reset it][decoy.Decoy.reset] after the test.

    This function is function-scoped [pytest fixture][] that will be
    automatically inserted by the plugin.

    [pytest fixture]: https://docs.pytest.org/en/latest/how-to/fixtures.html

    Example:
        ```python
        def test_my_thing(decoy: Decoy) -> None:
            my_fake_func = decoy.mock()
            # ...
        ```
    """
    decoy = Decoy()
    yield decoy
    decoy.reset()



"""Spy activity log."""
from typing import List, Sequence

from .errors import MissingRehearsalError
from .spy_events import (
    AnySpyEvent,
    SpyCall,
    SpyEvent,
    SpyPropAccess,
    WhenRehearsal,
    VerifyRehearsal,
    PropAccessType,
    PropRehearsal,
)


class SpyLog:
    """Log of all Spy activities in the Decoy container."""

    def __init__(self) -> None:
        self._log: List[AnySpyEvent] = []

    def push(self, spy_call: AnySpyEvent) -> None:
        """Add a new spy call to the stack."""
        self._log.append(spy_call)

    def consume_when_rehearsal(self, ignore_extra_args: bool) -> WhenRehearsal:
        """Consume the last call to a Spy as a `when` rehearsal.

        This marks a call as a rehearsal but does not remove it from the stack.
        """
        try:
            event = self._log[-1]
        except IndexError:
            raise MissingRehearsalError()
        if not isinstance(event, SpyEvent):
            raise MissingRehearsalError()

        spy, payload = _apply_ignore_extra_args(event, ignore_extra_args)

        rehearsal = WhenRehearsal(spy=spy, payload=payload)
        self._log[-1] = rehearsal
        return rehearsal

    def consume_verify_rehearsals(
        self,
        count: int,
        ignore_extra_args: bool,
    ) -> List[VerifyRehearsal]:
        """Consume the last `count` calls to Spies as rehearsals.

        This marks calls as rehearsals but does not remove them from the stack.
        """
        # events = self._log[-count:]
        rehearsals: List[VerifyRehearsal] = []
        index = len(self._log) - 1

        while len(rehearsals) < count:
            if index < 0:
                raise MissingRehearsalError()

            event = self._log[index]

            if not isinstance(event, (SpyEvent, PropRehearsal)):
                raise MissingRehearsalError()

            if _is_verifiable(event):
                rehearsal = VerifyRehearsal(
                    *_apply_ignore_extra_args(event, ignore_extra_args)
                )
                rehearsals.append(rehearsal)
                self._log[index] = rehearsal

            index = index - 1

        return list(reversed(rehearsals))

    def consume_prop_rehearsal(self) -> PropRehearsal:
        """Consume the last property get as a rehearsal."""
        try:
            event = self._log[-1]
        except IndexError:
            raise MissingRehearsalError()

        spy, payload = event

        if (
            not isinstance(event, SpyEvent)
            or not isinstance(payload, SpyPropAccess)
            or payload.access_type != PropAccessType.GET
        ):
            raise MissingRehearsalError()

        rehearsal = PropRehearsal(spy, payload)
        self._log[-1] = rehearsal
        return rehearsal

    def get_calls_to_verify(self, spy_ids: Sequence[int]) -> List[SpyEvent]:
        """Get all non-rehearsal calls to the spies in the given rehearsals."""
        return [
            event
            for event in self._log
            if event.spy.id in spy_ids
            and isinstance(event, SpyEvent)
            and _is_verifiable(event)
        ]

    def get_all(self) -> List[AnySpyEvent]:
        """Get a list of all calls and rehearsals made."""
        return list(self._log)

    def clear(self) -> None:
        """Remove all stored calls."""
        self._log.clear()


def _is_verifiable(event: AnySpyEvent) -> bool:
    return isinstance(event, SpyEvent) and (
        isinstance(event.payload, SpyCall)
        or event.payload.access_type != PropAccessType.GET
    )


def _apply_ignore_extra_args(event: AnySpyEvent, ignore_extra_args: bool) -> SpyEvent:
    spy, payload = event

    if isinstance(payload, SpyCall):
        payload = payload._replace(ignore_extra_args=ignore_extra_args)

    return SpyEvent(spy=spy, payload=payload)



"""Decoy implementation logic."""
import inspect
from typing import Any, Callable, Optional

from .call_handler import CallHandler
from .errors import MockNotAsyncError
from .spy import SpyCreator
from .spy_events import WhenRehearsal, PropAccessType, SpyEvent, SpyInfo, SpyPropAccess
from .spy_log import SpyLog
from .stub_store import StubBehavior, StubStore
from .types import ContextValueT, ReturnT
from .verifier import Verifier
from .warning_checker import WarningChecker

# ensure decoy.core does not pollute Pytest tracebacks
__tracebackhide__ = True


class DecoyCore:
    """The DecoyCore class implements the main logic of Decoy."""

    def __init__(
        self,
        verifier: Optional[Verifier] = None,
        warning_checker: Optional[WarningChecker] = None,
        stub_store: Optional[StubStore] = None,
        spy_log: Optional[SpyLog] = None,
        call_handler: Optional[CallHandler] = None,
        spy_creator: Optional[SpyCreator] = None,
    ) -> None:
        """Initialize the DecoyCore with its dependencies."""
        self._verifier = verifier or Verifier()
        self._warning_checker = warning_checker or WarningChecker()
        self._stub_store = stub_store or StubStore()
        self._spy_log = spy_log or SpyLog()
        self._call_hander = call_handler or CallHandler(
            spy_log=self._spy_log,
            stub_store=self._stub_store,
        )
        self._spy_creator = spy_creator or SpyCreator(call_handler=self._call_hander)

    def mock(
        self,
        *,
        spec: Optional[Any] = None,
        name: Optional[str] = None,
        is_async: bool = False,
    ) -> Any:
        """Create and register a new spy."""
        return self._spy_creator.create(spec=spec, name=name, is_async=is_async)

    def when(self, _rehearsal: ReturnT, *, ignore_extra_args: bool) -> "StubCore":
        """Create a new stub from the last spy rehearsal."""
        rehearsal = self._spy_log.consume_when_rehearsal(
            ignore_extra_args=ignore_extra_args
        )
        return StubCore(rehearsal=rehearsal, stub_store=self._stub_store)

    def verify(
        self,
        *_rehearsals: ReturnT,
        times: Optional[int],
        ignore_extra_args: bool,
    ) -> None:
        """Verify that a Spy or Spies were called."""
        rehearsals = self._spy_log.consume_verify_rehearsals(
            count=len(_rehearsals),
            ignore_extra_args=ignore_extra_args,
        )
        calls = self._spy_log.get_calls_to_verify([r.spy.id for r in rehearsals])

        self._verifier.verify(rehearsals=rehearsals, calls=calls, times=times)

    def prop(self, _rehearsal: ReturnT) -> "PropCore":
        """Get a property setter/deleter rehearser."""
        spy, payload = self._spy_log.consume_prop_rehearsal()

        return PropCore(spy=spy, prop_name=payload.prop_name, spy_log=self._spy_log)

    def reset(self) -> None:
        """Reset and remove all stored spies and stubs."""
        calls = self._spy_log.get_all()
        self._warning_checker.check(calls)
        self._spy_log.clear()
        self._stub_store.clear()


class StubCore:
    """The StubCore class implements the main logic of a Decoy Stub."""

    def __init__(self, rehearsal: WhenRehearsal, stub_store: StubStore) -> None:
        """Initialize the Stub with a configuration."""
        self._rehearsal = rehearsal
        self._stub_store = stub_store

    def then_return(self, *values: ReturnT) -> None:
        """Set the stub to return value(s)."""
        for i, return_value in reversed(list(enumerate(values))):
            self._stub_store.add(
                rehearsal=self._rehearsal,
                behavior=StubBehavior(
                    return_value=return_value,
                    once=(i != len(values) - 1),
                ),
            )

    def then_raise(self, error: Exception) -> None:
        """Set the stub to raise an error."""
        self._stub_store.add(
            rehearsal=self._rehearsal,
            behavior=StubBehavior(error=error),
        )

    def then_do(self, action: Callable[..., ReturnT]) -> None:
        """Set the stub to perform an action."""
        spy_info = self._rehearsal.spy

        if inspect.iscoroutinefunction(action) and not spy_info.is_async:
            raise MockNotAsyncError(
                f"Cannot configure {spy_info.name} to call {action}"
                f" because {spy_info.name} is not asynchronous."
            )

        self._stub_store.add(
            rehearsal=self._rehearsal,
            behavior=StubBehavior(action=action),
        )

    def then_enter_with(self, value: ContextValueT) -> None:
        """Set the stub to return a ContextManager wrapped value."""
        self._stub_store.add(
            rehearsal=self._rehearsal,
            behavior=StubBehavior(context_value=value),
        )


class PropCore:
    """Main logic of a property access rehearser."""

    def __init__(
        self,
        spy: SpyInfo,
        prop_name: str,
        spy_log: SpyLog,
    ) -> None:
        self._spy = spy
        self._prop_name = prop_name
        self._spy_log = spy_log

    def set(self, value: Any) -> None:
        """Create a property setter rehearsal."""
        event = SpyEvent(
            spy=self._spy,
            payload=SpyPropAccess(
                prop_name=self._prop_name,
                access_type=PropAccessType.SET,
                value=value,
            ),
        )
        self._spy_log.push(event)

    def delete(self) -> None:
        """Create a property deleter rehearsal."""
        event = SpyEvent(
            spy=self._spy,
            payload=SpyPropAccess(
                prop_name=self._prop_name,
                access_type=PropAccessType.DELETE,
            ),
        )
        self._spy_log.push(event)
